@@ -2,13 +2,15 @@ import argparse
 import curses
 import datetime
 import minorimpact
-#import logging
+import logging
 import os
 import os.path
 import re
 import subprocess
 import sys
 
+from minorimpact.curses import getstring, highlight
+
 import zlink
 import zlink.globalvars
 
@@ -39,9 +41,9 @@ class Note():
 
         self.parsed = self.parsefile()
 
+        self.backlinks = self.parselinks("backlinks")
         self.default = self.parsed['default']
         self.links = self.parselinks()
-        self.backlinks = self.parselinks("backlinks")
         self.references = self.parsereferences()
 
     def __str__(self):
@@ -49,7 +51,7 @@ class Note():
         #if (self.id is not None):
         #    str = str + f"[_metadata_:id]:- {self.id}\n"
         if (len(self.tags) > 0):
-            str = str + f"[_metadata_:tags]:- " + ",".join(self.tags) + "\n"
+            str = str + f'[_metadata_:tags]:- "' + ','.join(self.tags) + '"\n'
 
         if (len(str) > 0): str = str + "\n"
         for i in self.default:
@@ -241,7 +243,7 @@ class Note():
         section = "default"
         for l in lines:
             # collect metadata
-            m = re.search("^\[_metadata_:(.+)\]:- +(.+)$", l)
+            m = re.search('^\[_metadata_:(.+)\]:- +"?([^"]+)"?$', l)
             if (m):
                 key = m.group(1)
                 value = m.group(2)
@@ -326,6 +328,9 @@ class Note():
         self.__init__(self.filename)
 
     def search(self, search_string):
+        # TODO: Make it so any string that starts with '#' will also match tags, even though
+        #       they don't have a 'hashtag' in their raw form.
+        search_string = search_string.lower()
         m = re.search(search_string, self.title.lower())
         if (m): return True
         m = re.search(search_string, self.id.lower())
@@ -333,6 +338,7 @@ class Note():
         for t in self.tags:
             m = re.search(search_string, t.lower())
             if (m): return True
+
         for l in self.default:
             m = re.search(search_string, l.lower())
             if (m): return True
@@ -340,13 +346,19 @@ class Note():
         for r in self.references:
             if (r.search(search_string) is True):
                 return True
+        return False
 
     # Change the order value of the current note.
     def updateorder(self, new_order):
         original_file = self.filename
         self.order = new_order
         self.filename = "{:04d} - {} - {}.md".format(self.order, self.id, self.title)
+        logging.debug(f"Moved {original_file} to {self.filename}")
         os.rename(original_file, self.filename)
+        files = loadnotes()
+        for f in files:
+            n = Note(f)
+            n.updatelinks(original_file, self.filename)
 
     def updatetags(self, new_tags):
         tags = new_tags.split(",")
@@ -356,31 +368,45 @@ class Note():
         self.write()
 
     def updatetitle(self, new_title):
+        # TODO: Should this just use the write() method?  Does that method know how handle changes that would result in an 
+        #       altered filename?  If it doesn't, it really, really should.
         original_file = self.filename
         self.title = new_title
         self.filename = "{:04d} - {} - {}.md".format(self.order, self.id, self.title)
         os.rename(original_file, self.filename)
 
+    # Change any links for this note from 'url' to 'new_url'.
     def updatelinks(self, url, new_url):
+        # TODO: Make this work with actual link objects, as opposed to just passing around ghetto urls.
         new_note = None
         if (new_url is not None):
             new_note = Note(new_url)
 
-        for i in self.links:
-            if (i.url == url):
+        for l in self.links:
+            if (l.url == url):
                 if (new_note is None):
-                    self.links.remove(i)
+                    self.links.remove(l)
                 else:
-                    i.url = new_note.filename
-                    i.text = new_note.title
+                    logging.debug('changing link from %s:"%s" to %s:"%s"', l.url, l.text, new_note.filename, new_note.title) 
+                    l.url = new_note.filename
+                    l.text = new_note.title
                 
-        for i in self.backlinks:
-            if (i.url == url):
+        for b in self.backlinks:
+            if (b.url == url):
                 if (new_note is None):
-                    self.backlinks.remove(i)
+                    self.backlinks.remove(b)
                 else:
-                    i.url = new_note.filename
-                    i.text = new_note.title
+                    logging.debug('changing backlink from %s:"%s" to %s:"%s"', b.url, b.text, new_note.filename, new_note.title) 
+                    b.url = new_note.filename
+                    b.text = new_note.title
+
+        for r in self.references:
+            if (r.link == url):
+                if (new_url is None):
+                    self.references.remove(r);
+                else:
+                    r.link = new_url
+
         self.write()
         
     def view(self, stdscr):
@@ -425,7 +451,7 @@ class Note():
             if (select is True):
                 #c = stdscr.inch(select_y, select_x)
                 #stdscr.insch(select_y, select_x, c, curses.A_REVERSE)
-                zlink.zlink.highlight(stdscr, select_y, select_x, mark_y, mark_x)
+                highlight(stdscr, select_y, select_x, mark_y, mark_x)
             stdscr.refresh()
             command = stdscr.getkey()
 
@@ -505,11 +531,12 @@ class Note():
                 continue
             elif (command == "f"):
                 filebrowser.browse(stdscr)
-            elif (command == "l"):
-                # Link a note to this note
+            elif (command == 'l'):
                 if (zlink.globalvars.link_note is None):
+                    # store this note for linking later
                     zlink.globalvars.link_note = self
                 else:
+                    # link the previous note to the current note
                     self.addnotelink(zlink.globalvars.link_note)
                     self.write()
                     zlink.globalvars.link_note.addnotebacklink(self)
@@ -543,7 +570,7 @@ class Note():
             elif (command == "\n"):
                 if (select is True):
                     if (mark_x is not None):
-                        text = zlink.zlink.highlight(stdscr, select_y, select_x, mark_y, mark_x)
+                        text = highlight(stdscr, select_y, select_x, mark_y, mark_x)
                         link = Link(self.filename, self.title)
                         zlink.globalvars.copy = Reference(link, text)
                         #pyperclip.copy(copy.__str__())
@@ -585,11 +612,13 @@ class Note():
                 stdscr.addstr(" c              - enter selection mode to copy text to save the clipboard as a reference\n")
                 stdscr.addstr(" e              - open this note in the external editor (set the EDITOR environment variable)\n")
                 stdscr.addstr(" l              - press once to set this note as the target.  Navigate to another note and press\n")
-                stdscr.addstr("                  'l' again to add a link to the first note from the second note.\n")
+                stdscr.addstr("                  'l' again to add a link to the target note to the current note\n")
                 stdscr.addstr(" p              - paste a reference from the clipboard to the current note\n")
                 stdscr.addstr(" q              - quit\n")
                 stdscr.addstr(" r              - rename note\n")
                 stdscr.addstr(" t              - edit tags\n")
+                stdscr.addstr(" <del> or       - delete the currently selected link or backlink\n")
+                stdscr.addstr(" <backspace>\n")
                 stdscr.addstr(" ?              - this help screen\n")
                 stdscr.addstr("\n")
                 stdscr.addstr("Selection Mode Commands\n\n", curses.A_BOLD)
@@ -611,6 +640,11 @@ class Note():
                 command = stdscr.getkey()
 
     def write(self):
+        # TODO: Make this compare the filename in the object with the filename that would be generated from the object
+        #       data, and if they don't match, you know... fucking fix it.  This shouldn't be happening in updateTitle() or 
+        #       updateOrder().
+        # NOTE: I think i may have looked into this before, but see if there's a way for an object to detect changes to itself and 
+        #       perform actions if something is different.  That could be a thing, right?
         with open(self.filename, "w") as f:
             f.write(self.__str__())
             f.close()
@@ -633,7 +667,6 @@ class Reference():
 
 class NoteBrowser():
     def browse(self, stdscr, filename=None):
-
         stdscr.clear()
 
         files = loadnotes()
@@ -655,7 +688,6 @@ class NoteBrowser():
 
             status = ""
 
-
             if (note1 is not None):
                 newnote = note1.view(stdscr)
                 if (zlink.globalvars.reload):
@@ -696,15 +728,23 @@ class NoteBrowser():
                     if (i < top): continue
                     if (i > (top + curses.LINES - 2 )): continue
                     f = files[i]
-                    max_width = curses.COLS - 2
+                    max_width = curses.COLS - 6
+                    menu_item = "{:" + str(max_width) + "." + str(max_width) + "s}\n"
+                    note = Note(f)
+                    
+                    menu_item = menu_item.format(f)
                     if (i == selected):
-                        stdscr.addstr(("{:" + str(max_width) + "." + str(max_width) + "s}\n").format(f), curses.A_REVERSE)
+                        stdscr.addstr(menu_item, curses.A_REVERSE)
                     else:
-                        stdscr.addstr(("{:" + str(max_width) + "." + str(max_width) + "s}\n").format(f))
+                        stdscr.addstr(menu_item)
                 status = f"{selected+1} of {len(files)}"
 
             if (move):
                 status = f"{status} MOVING"
+            if (zlink.globalvars.link_note is not None):
+                status = f"{status} LINKING"
+            if (zlink.globalvars.filter != ""):
+                status = f"{status} FILTERED:'{zlink.globalvars.filter}'"
 
             if (status):
                 # Make sure a long status doesn't push 
@@ -714,20 +754,24 @@ class NoteBrowser():
             stdscr.refresh()
             command = stdscr.getkey()
 
-            if (command == "KEY_DOWN" or command == "KEY_RIGHT"):
-                original_selected = selected
-                selected += 1
-                if (selected > len(files)-1):
-                    selected = 0
-                if (move is True):
-                    files = swapnotes(files, original_selected, selected)
-            elif (command == "KEY_UP" or command == "KEY_LEFT"):
+            if (command == "KEY_UP"):
                 original_selected = selected
                 selected -= 1
                 if (selected < 0):
                     selected = len(files)-1
                 if (move is True):
                     files = swapnotes(files, original_selected,selected)
+            elif (command == "KEY_DOWN"):
+                original_selected = selected
+                selected += 1
+                if (selected > len(files)-1):
+                    selected = 0
+                if (move is True):
+                    files = swapnotes(files, original_selected, selected)
+            #elif (command == "KEY_LEFT"):
+            #    note = Note(files[selected])
+            #elif (command == "KEY_RIGHT"):
+            #    note = Note(files[selected])
             elif (command == "KEY_END" or command == "G"):
                 # TODO: Does pgup/pgdown/home/end have to kill the "move" command? or does
                 #   it make sense for me to be able to move a note a vast difference, rather than
@@ -752,49 +796,37 @@ class NoteBrowser():
                 selected -= curses.LINES - 2
                 if (selected < 0):
                     selected = 0
-            elif (command == "a"):
+            elif (command == 'a' or command == 'o'):
+                if (zlink.globalvars.filter != ""):
+                    continue
                 move = False
                 new_title = getstring(stdscr, "New Note: ", 80)
                 if (new_title == ""):
                     continue
-                # based on the selected note, figure out how many notes we have to adjust to make a hole
-                if (len(files) == 0):
-                    next_order = 1
-                elif (selected < len(files)-1):
-                    note = Note(files[selected+1])
-                    next_order = note.order + 1
-                    for f in files[selected:]:
-                        n = Note(f)
-                        if (n.order > next_order):
-                            break
-                        next_order = n.order + 1
-
-                    # now that we have the first free spot, move everything up one
-                    tmp_files = files[selected+1:]
-                    tmp_files.reverse()
-                    for f in tmp_files:
-                        n = Note(f)
-                        if (n.order < next_order):
-                            original_file = n.filename
-                            n.updateorder(next_order)
-                            files = loadnotes()
-                            for f2 in files:
-                                n2 = Note(f2)
-                                n2.updatelinks(original_file, n.filename)
-                            next_order -= 1
-                else:
-                    note = Note(files[-1])
-                    next_order = note.order + 1
-                today = datetime.datetime.now()
-                date = today.strftime("%Y-%m-%d %H-%M")
-                filename = "{:04d} - {} - {}.md".format(next_order, date, new_title)
-                new_note = Note(filename)
-                new_note.write()
+                new_order = makehole(files, selected+1)
+                new_note = newNote(new_order, new_title)
                 files = loadnotes()
                 note1 = new_note
                 selected = files.index(note1.filename)
-            elif (command == "KEY_DC" or command == "d"):
+            elif (command == 'A' or command == 'O'):
+                if (zlink.globalvars.filter != ""):
+                    continue
                 move = False
+                new_title = getstring(stdscr, "New Note: ", 80)
+                if (new_title == ""):
+                    continue
+                new_order = makehole(files, selected-1)
+                new_note = newNote(new_order, new_title)
+                files = loadnotes()
+                note1 = new_note
+                selected = files.index(note1.filename)
+            elif (command == "KEY_DC" or command == 'd' or command == '^?'):
+                if (zlink.globalvars.filter != ""):
+                    continue
+                if (move is True or zlink.globalvars.link_note is not None):
+                    move = False
+                    zlink.globalvars.link_note = None
+                    continue
                 note = Note(files[selected])
                 original_file = note.filename
                 confirm = getstring(stdscr, "Are you sure you want to delete this note? (y/N):", 1)
@@ -807,14 +839,60 @@ class NoteBrowser():
             elif (command == "f"):
                 #f = FileBrowser()
                 filebrowser.browse(stdscr)
-            elif (command == "m"):
+            elif (command == 'F'):
+                original_selected = selected
+                note = Note(files[selected])
+                new_filter = getstring(stdscr, "filter for: ").lower()
+                if (new_filter != ""):
+                    zlink.globalvars.filter = new_filter
+                if (zlink.globalvars.filter == ""):
+                    continue
+
+                move = False
+                zlink.globalvars.link_note = None
+                files = loadnotes()
+                if (len(files) == 0):
+                    zlink.globalvars.filter = ""
+                    files = loadnotes()
+                    selected = files.index(note.filename)
+                    status = f"{status} NOT FOUND"
+                    continue
+                try:
+                    selected = files.index(note.filename)
+                except ValueError:
+                    selected = 0
+
+            elif (command == "l"):
+                move = False
+                note = Note(files[selected])
+                if (zlink.globalvars.link_note is None):
+                    # store this note for linking later
+                    zlink.globalvars.link_note = note
+                elif (note is not None):
+                    # link the previous note to the current note
+                    note.addnotelink(zlink.globalvars.link_note)
+                    note.write()
+                    zlink.globalvars.link_note.addnotebacklink(note)
+                    zlink.globalvars.link_note.write()
+                    zlink.globalvars.link_note = None
+            elif (command == 'm'):
+                # TODO: moving a note down from the last position to reposition it back a the top ends up swapping the first and last item...
+                #       I think that's wrong, but it's not immediately clear what the desired behavior would be.  Would it just become the new first item
+                #       and force everything else to move down?  Or would it become the *second* item, because it would still technically be swapping
+                #       with the item below it, and we want to the current first item to stay first?  It seems like the more intuitive method would be to
+                #       treat the first and last items as connected, but also anchored to their respective positions in the list.  So moving the last item
+                #       'down' would move it to the number 2 slot, and moving the first item 'up' would move it to the max-1 spot.
+                if(zlink.globalvars.filter != ""):
+                    continue
                 if (move is True):
                     move = False
                 else:
                     move = True
-            elif (command == "/"):
+                    zlink.globalvars.link_note = None
+            elif (command == '/'):
                 original_selected = selected
-                move = False
+                #move = False
+                #zlink.globalvars.link_note = None
                 new_search = getstring(stdscr, "Search for: ")
                 if (new_search != ""):
                     search = new_search
@@ -836,24 +914,54 @@ class NoteBrowser():
                         selected = files.index(f)
                         break
             elif (command == "\n"):
-                move = False
+                if (move is True or zlink.globalvars.link_note is not None):
+                    # clear any 'special' modes.
+                    move = False
+                    note = Note(files[selected])
+                    if (note is not None):
+                        if (zlink.globalvars.link_note is not None):
+                            # link the previous note to the current note
+                            note.addnotelink(zlink.globalvars.link_note)
+                            note.write()
+                            zlink.globalvars.link_note.addnotebacklink(note)
+                            zlink.globalvars.link_note.write()
+                            zlink.globalvars.link_note = None
+                    continue
+                    
                 note1 = Note(files[selected])
                 #selected = 0
                 #top = 0
-            elif (command == ""):
-                move = False
+            elif (command == ''):
+                # clear any 'special' modes.
+                if (move == True or zlink.globalvars.link_note != None):
+                    move = False
+                    zlink.globalvars.link_note = None
+                elif (zlink.globalvars.filter != ""):
+                    zlink.globalvars.filter = ""
+                    f = files[selected]
+                    files = loadnotes()
+                    try:
+                        selected = files.index(f)
+                    except ValueError:
+                        selected = 0
             elif (command == "?"):
                 stdscr.clear()
                 stdscr.addstr("Editing Commands\n", curses.A_BOLD)
-                stdscr.addstr("a                - add a new note after the selected note\n")
-                stdscr.addstr("d or <del>       - delete the currently selected note\n")
-                stdscr.addstr("m                - change to 'move' mode.  <up>/<down> will move the selected note. <esc> to cancel\n")
+                stdscr.addstr("A or O           - add a new note before the selected note.*\n")
+                stdscr.addstr("a or o           - add a new note after the selected note.*\n")
+                stdscr.addstr("d or <del>       - delete the currently selected note.*\n")
+                stdscr.addstr("l                - will set the current note to the target and activate 'link' mode.  Navigating to any other\n")
+                stdscr.addstr("                   note and pressing 'l' (or <enter>) again will link the current note to the target note.\n")
+                stdscr.addstr("m                - change to 'move' mode.  <up>/<down> will move the selected note. <esc> to cancel.*\n")
                 stdscr.addstr("q                - quit\n")
-                stdscr.addstr("/                - enter a string to search for\n")
                 stdscr.addstr("?                - this help screen\n")
 
                 stdscr.addstr("\n")
                 stdscr.addstr("Navigation Commands\n", curses.A_BOLD)
+                stdscr.addstr("F                - create a filter that limits visible notes to only those that match the entered string. <esc>\n")
+                stdscr.addstr("                   to cancel.\n")
+                stdscr.addstr("/                - enter a string and press <enter> to jump to the next matching record.  Uses the same \n")
+                stdscr.addstr("                   syntax as filter.\n")
                 stdscr.addstr("f                - open the file browser\n")
                 stdscr.addstr("<home>           - first note\n")
                 stdscr.addstr("<up>             - previous/next note\n")
@@ -861,8 +969,18 @@ class NoteBrowser():
                 stdscr.addstr("<pgdown> or ^d   - move the curser up one screen\n")
                 stdscr.addstr("<down>           - next note\n")
                 stdscr.addstr("<end> or G       - last note\n")
-                stdscr.addstr("<enter>          - open the selected note\n")
-                stdscr.addstr("<esc>            - cancel 'move' mode, link mode")
+                stdscr.addstr("<enter>          - open the selected note (or turn off 'move' or 'link'  mode)\n")
+                stdscr.addstr("<esc>            - cancel 'move' mode, 'link' mode, clear the current filter")
+                stdscr.addstr("\n")
+                stdscr.addstr("\n")
+                stdscr.addstr("* these commands do NOT work while a filter is engaged\n")
+                # NOTE: the problem is that all of these commands basically just walk through the list and do their things
+                #       based on what they find.  If that list is incomplete, a lot of things (like pruning links to deleted notes
+                #       or shuffling all of the existing notes to make room for a new one) won't happen for any of the filtered 
+                #       filtered results.  I need to make the global list some kind of dynamic object (rather than a simple array)
+                #       or add a flag to "loadnotes()" that can tell it to ignore global filter when it runs, and these commands can
+                #       all work on a local copy.  Or I can clear the filter, load the results, process them, and then put the filter
+                #       back... it's not hard, I'm just not going to do it now.
 
                 stdscr.addstr(curses.LINES-1,0,"Press any key to continue", curses.A_BOLD)
                 stdscr.refresh()
@@ -874,8 +992,8 @@ def gethole(files, position=0):
     if (len(files) == 0):
         next_order = 1
     elif (position < len(files)-1):
-        next_order = note.order + 1
         note = Note(files[position])
+        next_order = note.order + 1
         for f in files[position:]:
             n = Note(f)
             if (n.order > next_order):
@@ -886,15 +1004,6 @@ def gethole(files, position=0):
          next_order = note.order + 1
     return next_order
 
-# Request a string from the user.
-def getstring(stdscr, prompt_string, maxlength=40):
-    curses.echo()
-    stdscr.addstr(curses.LINES-1, 0, prompt_string)
-    stdscr.refresh()
-    input = stdscr.getstr(curses.LINES-1, len(prompt_string), maxlength).decode(encoding='utf-8')
-    curses.noecho()
-    return input
-
 # Return the item at the 'top' of the screen, based on what is currently selected.
 def gettop(selected, current_top, maxlength, center=False):
     top = current_top
@@ -911,24 +1020,73 @@ def gettop(selected, current_top, maxlength, center=False):
 
 # Read the list of notes from the disk.
 def loadnotes():
-    files = [f for f in os.listdir(".") if(os.path.isfile(os.path.join(".", f)) and re.search("^\d+ - .+\.md$",f))]
+    files = []
+    for f in os.listdir("."):
+        if (os.path.isfile(os.path.join(".", f)) and re.search("^\d+ - .+\.md$",f)):
+            if (zlink.globalvars.filter != ""):
+                note = Note(f)
+                logging.debug("filtering for %s", zlink.globalvars.filter)
+                if (note.search(zlink.globalvars.filter) == False):
+                    continue
+                logging.debug("  found")
+            files.append(f)
+    #files = [f for f in os.listdir(".") if(os.path.isfile(os.path.join(".", f)) and re.search("^\d+ - .+\.md$",f))]
     files.sort()
     return files
 
+def makehole(files, position):
+    files = loadnotes()
+    note = Note(files[position])
+    hole = 1
+    if (position < 0):
+        position = 0
+    else:
+        if (position > len(files)):
+            position = len(files)
+        previous_note = Note(files[position-1])
+        hole = previous_note.order+1
+
+    last_order = hole
+    logging.debug(f"making a hole at {hole}({position})")
+    for i in range(position,len(files)):
+        logging.debug(f"evaluating postion {i}")
+        n = None
+        try:
+            n = Note(files[i])
+        except:
+            raise Exception(f"Can't open '{files[i]}'")
+
+        logging.debug(f"  {n.order}:{n.filename}")
+        if (n.order <= last_order):
+            original_file = n.filename
+            n.updateorder(last_order + 1)
+            #files[i] = n.filename
+            last_order = last_order + 1
+
+    files = loadnotes()
+    return hole
+
 def swapnotes(files, original_pos, new_pos):
     n1 = Note(files[original_pos])
-    n1_file = n1.filename
-
+    n1_order = n1.order
     n2 = Note(files[new_pos])
-    n2_file = n2.filename
-    new_order = n1.order
-    if (n2.order == n1.order and new_pos < original_pos):
-        new_order = gethole(files, new_pos)
-    n1.updateorder(n2.order)
-    n2.updateorder(new_order)
+    n2_order = n2.order
+    if (n2_order == n1_order):
+        if (new_pos < original_pos):
+            n2.updateorder(makehole(files, original_pos))
+        elif (new_pos > original_pos):
+            n1.updateorder(makehole(files, new_pos))
+    else:
+        n1.updateorder(n2_order)
+        n2.updateorder(n1_order)
     files = loadnotes()
-    for f in files:
-        note = Note(f)
-        note.updatelinks(n1_file, n1.filename)
-        note.updatelinks(n2_file, n2.filename)
     return files
+
+def newNote(order, title):
+    today = datetime.datetime.now()
+    date = today.strftime("%Y-%m-%d %H-%M")
+    filename = "{:04d} - {} - {}.md".format(order, date, title)
+    new_note = Note(filename)
+    new_note.write()
+    return new_note
+
